/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 28/02/2008
   derniere modification : 29/02/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/interface/SCEFrustum.h>

static SCE_TMatrix4 clip;       /* matrices de projection &
                                   visualisation combinees */
static SCE_SBoundingBox *box = NULL;
static SCE_TVector3 points[8];  /* points temporaires d'une bounding box */
static int i = 0, j = 0;        /* compteurs */

/* ajoute le 28/02/2008 */
void SCE_Frustum_Init (SCE_SFrustum *f)
{
    for (i=0; i<6; i++)
        SCE_Plane_Init (&f->planes[i]);
}

/* ajoute le 28/02/2008 */
void SCE_Frustum_MakeFromMatrices (SCE_SFrustum *f, SCE_TMatrix4 view,
                                   SCE_TMatrix4 proj)
{
    SCE_Matrix4_Mul (proj, view, clip);
    /* extrait le plan de droite */
    SCE_Plane_Set (&f->planes[0],
                   clip[12]-clip[0], clip[13]-clip[1],
                   clip[14]-clip[2], clip[15]-clip[3]);
    /* extrait le plan de gauche */
    SCE_Plane_Set (&f->planes[1],
                   clip[12]+clip[0], clip[13]+clip[1],
                   clip[14]+clip[2], clip[15]+clip[3]);
    /* extrait le plan du bas */
    SCE_Plane_Set (&f->planes[2],
                   clip[12]+clip[4], clip[13]+clip[5],
                   clip[14]+clip[6], clip[15]+clip[7]);
    /* extrait le plan du haut */
    SCE_Plane_Set (&f->planes[3],
                   clip[12]-clip[4], clip[13]-clip[5],
                   clip[14]-clip[6], clip[15]-clip[7]);
    /* extrait le plan eloigne */
    SCE_Plane_Set (&f->planes[4],
                   clip[12]-clip[8], clip[13]-clip[9],
                   clip[14]-clip[10], clip[15]-clip[11]);
    /* extrait le plan proche */
    SCE_Plane_Set (&f->planes[5],
                   clip[12]+clip[8], clip[13]+clip[9],
                   clip[14]+clip[10], clip[15]+clip[11]);

#if 0
    /* extrait le plan de droite */
    SCE_Plane_Set (&f->planes[0], clip[3]-clip[0], clip[7]-clip[4],
                   clip[11]-clip[8], clip[15]-clip[12]);
    /* extrait le plan de gauche */
    SCE_Plane_Set (&f->planes[1], clip[3]+clip[0], clip[7]+clip[4],
                   clip[11]+clip[8], clip[15]+clip[12]);
    /* extrait le plan du bas */
    SCE_Plane_Set (&f->planes[2], clip[3]+clip[1], clip[7]+clip[5],
                   clip[11]+clip[9], clip[15]+clip[13]);
    /* extrait le plan du haut */
    SCE_Plane_Set (&f->planes[3], clip[3]-clip[1], clip[7]-clip[5],
                   clip[11]-clip[9], clip[15]-clip[13]);
    /* extrait le plan eloigne */
    SCE_Plane_Set (&f->planes[4], clip[3]-clip[2], clip[7]-clip[6],
                   clip[11]-clip[10], clip[15]-clip[14]);
    /* extrait le plan proche */
    SCE_Plane_Set (&f->planes[5], clip[3]+clip[2], clip[7]+clip[6],
                   clip[11]+clip[10], clip[15]+clip[14]);
#endif
    /* normalisation */
    for (i=0; i<6; i++)
        SCE_Plane_Normalize (&f->planes[i], SCE_TRUE);
}

/* ajoute le 28/02/2008 */
int SCE_Frustum_PointIn (SCE_SFrustum *f, float x, float y, float z)
{
    for (i=0; i<6; i++)
    {
        if (SCE_Plane_DistanceToPoint (&f->planes[i], x, y, z) <= 0.0f)
            return SCE_FALSE;
    }
    return SCE_TRUE;
}
int SCE_Frustum_PointInv (SCE_SFrustum *f, SCE_TVector3 p)
{
    for (i=0; i<6; i++)
    {
        if (SCE_Plane_DistanceToPointv (&f->planes[i], p) <= 0.0f)
            return SCE_FALSE;
    }
    return SCE_TRUE;
}

/* ajoute le 28/02/2008 */
int SCE_Frustum_BoundingBoxIn (SCE_SFrustum *f, SCE_SBoundingBox *b)
{
    int k = 0, total = 0;
    SCE_BoundingBox_GetPointsv (b, points);
    /* pour chaque plan */
    for (i=0; i<6; i++)
    {
        /* au moins un point doit etre a l'interieur */
        for (j=0; j<8; j++)
        {
            if (SCE_Plane_DistanceToPointv (&f->planes[i], points[j]) > 0.0f)
                k++;
        }
        if (k == 8) /* tous les points sont dedans */
            total++;
        else if (!k) /* aucun point n'est dedans, arret */
            return SCE_FRUSTUM_OUT;
        k = 0;
    }
    if (total == 6)
        return SCE_FRUSTUM_IN; /* entierement dedans */
    else
        return SCE_FRUSTUM_PARTIALLY; /* partiellement dedans */
}

/* ajoute le 28/02/2008 */
int SCE_Frustum_CubeIn (SCE_SFrustum *f, SCE_TVector3 o,
                        float w, float h, float d)
{
    SCE_BoundingBox_Set (box, o, w, h, d);
    return SCE_Frustum_BoundingBoxIn (f, box);
}
int SCE_Frustum_CubeInv (SCE_SFrustum *f, SCE_TVector3 o, SCE_TVector3 d)
{
    SCE_BoundingBox_Setv (box, o, d);
    return SCE_Frustum_BoundingBoxIn (f, box);
}
