/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2008
   derniere modification le 11/07/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECore.h>
#include <SCE/interface/SCEScene.h>

#include <SCE/utils/SCEVector.h>
#include <SCE/interface/SCEBoundingBox.h>

int SCE_Init_Scene (void)
{
    return SCE_OK;
}
void SCE_Quit_Scene (void)
{
}


/* ajoute le 10/07/2008 */
static void SCE_Scene_InitRenderStates (SCE_SSceneRenderStates *states)
{
    /* parametres par defaut des scenes */
    states->frustum_culling = SCE_FALSE;
    states->use_skybox = SCE_FALSE;
    states->lighting = SCE_TRUE;
}

/* ajoute le 03/04/2008 */
/* revise le 10/07/2008 */
static void SCE_Scene_InitRenderInfos (SCE_SSceneRenderInfos *infos)
{
    infos->rendertarget = NULL;
    infos->cubeface = 0;
    infos->camera = NULL;
    infos->models = NULL;
    SCE_Scene_InitRenderStates (&infos->states);
}

/* ajoute le 23/02/2008 */
/* revise le 06/04/2008 */
static void SCE_Scene_Init (SCE_SScene *scene)
{
    int i;
    scene->rootnode = NULL;
    scene->models = NULL;
    scene->lights = NULL;
    for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
        SCE_Scene_InitRenderInfos (&scene->render[i]);
    scene->ri = 0;
}

static void SCE_Scene_DeleteModelRenderInfos (void *infos)
{
    SCE_free (infos);
}

/* ajoute le 23/02/2008 */
/* revise le 20/03/2008 */
SCE_SScene* SCE_Scene_Create (void)
{
    int i;
    SCE_SScene *scene = NULL;
    SCE_btstart ();
    if (!(scene = SCE_malloc (sizeof *scene)))
        Logger_LogSrc ();
    else
    {
        SCE_Scene_Init (scene);
#define SCE_SCENE_ASSERT(c)\
        if (c)\
        {\
            SCE_Scene_Delete (scene);\
            Logger_LogSrc ();\
            SCE_btend ();\
            return NULL;\
        }

        SCE_SCENE_ASSERT (!(scene->models = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->lights = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->rootnode = SCE_Node_Create ()))
        for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
        {
            scene->render[i].models =
                SCE_List_Create (SCE_Scene_DeleteModelRenderInfos);
            SCE_SCENE_ASSERT (!(scene->render[i].models))
        }
#undef SCE_SCENE_ASSERT
    }
    SCE_btend ();
    return scene;
}
/* ajoute le 23/02/2008 */
void SCE_Scene_Delete (SCE_SScene *scene)
{
    if (scene)
    {
        int i;
        for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
            SCE_List_Delete (scene->render[i].models);
        SCE_Node_Delete (scene->rootnode);
        SCE_List_Delete (scene->lights);
        SCE_List_Delete (scene->models);
        SCE_free (scene);
    }
}


/* ajoute le 20/03/2008 */
void SCE_Scene_GenerateBoundingsBoxes (SCE_SScene *scene)
{
    SCE_SListIterator *it = NULL;
    SCE_List_ForEach (it, scene->models)
        SCE_Model_GenerateBoundingBox (SCE_List_GetData (it));
}


/* ajoute le 26/02/2008 */
SCE_SNode* SCE_Scene_GetRootNode (SCE_SScene *scene)
{
    return scene->rootnode;
}

/* ajoute le 26/02/2008 */
int SCE_Scene_AddNode (SCE_SScene *scene, SCE_SNode *node)
{
    if (SCE_Node_Attach (scene->rootnode, node) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 26/02/2008 */
/* revise le 08/07/2008 */
int SCE_Scene_AddModel (SCE_SScene *scene, SCE_SModel *model)
{
    int i;

    /* ajout du modele a la liste des modeles */
    if (SCE_List_AddBack (scene->models, model) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }

    /* ajout des elements aux listes */
    for (i=0; i<SCE_MAX_RENDER_INFOS; i++)
    {
        SCE_SSceneModelRenderInfos *infos = SCE_malloc (sizeof *infos);
        if (!infos)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        infos->model = model;
        infos->num_lights = 0;
        /* ajout de l'infos de rendu pour le modele */
        if (SCE_List_AddBack (scene->render[i].models, infos) < 0)
        {
            Logger_LogSrc ();
            SCE_free (infos); /* TODO: ne frees pas les precedents... osef */
            return SCE_ERROR;
        }
    }
    return SCE_OK;
}

/* ajoute le 06/04/2008 */
/* revise le 07/04/2008 */
int SCE_Scene_AddSkybox (SCE_SScene *scene, SCE_SModel *skybox)
{
    if (SCE_List_AddFront (scene->models, skybox) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    if (!SCE_Model_HaveRenderFunc (skybox))
        SCE_Model_SetRenderFunc (skybox, SCE_Scene_RenderSkybox, scene);
    return SCE_OK;
}

/* ajoute le 20/03/2008 */
int SCE_Scene_AddLight (SCE_SScene *scene, SCE_SLight *light)
{
    if (SCE_List_AddBack (scene->lights, light) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}



/* ajoute le 03/04/2008 */
/* ajoute le 06/04/2008 */
/* TODO: cette func aussi elle pue */
void SCE_Scene_RenderSkybox (SCE_SModel *mdl, void *s)
{
    SCE_SScene *scene = s;
    SCE_SCamera *cam = scene->render[scene->ri].camera;
    SCE_TVector3 campos = {0., 0., 0.};
    SCE_TMatrix4 cmat;
    float *mat = SCE_Node_GetMatrix (SCE_Model_GetNode (mdl));

    SCE_Matrix4_Copy (cmat, SCE_Camera_GetView (cam));
    SCE_Matrix4_Inverse (cmat);
    SCE_Matrix4_GetTranslation (cmat, campos);
    SCE_Matrix4_Copy (cmat, mat); /* sauvegarde de la matrice */
    SCE_Matrix4_MulTranslatev (mat, campos);
    /* glPushAttribs */
    SCE_CSetState (GL_DEPTH_TEST, SCE_FALSE);
    glDepthMask (0); /* TODO: hack laid */
    SCE_Model_RenderFunc (mdl, NULL);
    glDepthMask (1);
    SCE_CSetState (GL_DEPTH_TEST, SCE_TRUE);
    SCE_Matrix4_Copy (mat, cmat); /* restauration de la matrice */
}


/* ajoute le 03/04/2008 */
/* TODO: da shit func */
void SCE_Scene_PrepareRendering (SCE_SScene *scene)
{
    if (scene->render[scene->ri].states.use_skybox)
        SCE_CClear (GL_DEPTH_BUFFER_BIT);
    else
        SCE_CClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}



/* ajoute le 05/04/2008 */
/* indique si un modele est dans le frustum de vue */
static int SCE_Scene_ModelInFrustum (SCE_SScene *scene, SCE_SModel *mdl)
{
    int render;
    SCE_SBoundingBox *box = SCE_Model_GetBoundingBox (mdl);
    SCE_SCamera *cam = scene->render[scene->ri].camera;

    /* application de la matrice a la bouding box */
    SCE_BoundingBox_Push (box, SCE_Node_GetFinalMatrix(SCE_Model_GetNode(mdl)));
    /* on verifie si le modele est dans le frustum */
    render = SCE_Frustum_BoundingBoxIn (SCE_Camera_GetFrustum (cam),
                                        SCE_Model_GetBoundingBox (mdl));
    SCE_BoundingBox_Pop (box);
    return render;
}

/* ajoute le 07/07/2008 */
/* construit la liste des modeles qui seront dessines */
static void SCE_Scene_SelectInFrustumModels (SCE_SScene *scene)
{
    SCE_SListIterator *i = NULL;
    SCE_SSceneModelRenderInfos *info = NULL;
    int ri = scene->ri;

    /* TODO: risque de faire des calculs doublons */
    SCE_Camera_UpdateFrustum (scene->render[ri].camera);

    /* pour chaque modele... */
    SCE_List_ForEach (i, scene->render[ri].models)
    {
        info = SCE_List_GetData (i);
        /* s'il a ete valide dans les (eventuels) precedents tests,
           on effectue ce nouveau test */
        if (info->draw)
        {
            /* ... on verifie s'il est dans le frustum de vue */
            if (!SCE_Scene_ModelInFrustum (scene, info->model))
                info->draw = SCE_FALSE;
        }
    }
}

/* ajoute le 11/07/2008 */
static void SCE_Scene_SelectActivedModels (SCE_SScene *scene)
{
    SCE_SListIterator *i = NULL;
    SCE_SSceneModelRenderInfos *info = NULL;
    int ri = scene->ri;

    /* pour chaque modele */
    SCE_List_ForEach (i, scene->render[ri].models)
    {
        info = SCE_List_GetData (i);
        /* on verifie s'il est active */
        if (SCE_Model_IsActived (info->model))
            info->draw = SCE_TRUE;
        else
            info->draw = SCE_FALSE;
    }
}

/* ajoute le 07/07/2008 */
/* toutes les lumieres dont infos->model est a portee sont enregistrees */
static void SCE_Scene_MakeLightsList (SCE_SScene *scene,
                                      SCE_SSceneModelRenderInfos *infos)
{
    SCE_SListIterator *i = NULL;
    unsigned int j = 0;
    SCE_SLight *light = NULL;
    SCE_SBoundingBox *box = NULL;
    SCE_TVector3 vec1, vec2;
    float length;
    float *p = NULL;
    SCE_SModel *mdl = infos->model;

    box = SCE_Model_GetBoundingBox (mdl);
    if (!box) /* TODO: ahem... */
        return;

    /* obtention de donnees necessaire au calcul des lumieres */
    SCE_BoundingBox_GetCenterv (box, vec1);
    SCE_Vector3_Copy (vec2, SCE_BoundingBox_GetPoints (box));
    length = SCE_Vector3_Distance (vec1, vec2);

    /* activation des lumieres */
    SCE_List_ForEach (i, scene->lights)
    {
        light = SCE_List_GetData (i);

        /* on verifie d'abord si la lumiere est activee */
        if (!SCE_Light_IsActived (light))
            continue;

        /* on n'active la lumiere que si le modele est a sa portee */
        p = SCE_Light_GetPosition (light);
        if (SCE_Vector3_Distance (p, vec1) <= SCE_Light_GetRadius(light)+length)
        {
            /* TODO: gerer les spot lights */
            /* ajout du modele a la liste des modeles */
            infos->lights[j] = light;
            j++;
            /* TODO: je ne considere que les premieres lumieres de la liste
               et non les plus importantes (cad les plus intenses/proches)
               penser a trier scene->lights, et ca sera cool */
            if (j >= SCE_MAX_LIGHTS_ON_MODEL)
                break;
        }
    }
    infos->num_lights = j;
}

/* ajoute le 08/07/2008 */
static void SCE_Scene_MakeLightsLists (SCE_SScene *scene)
{
    SCE_SSceneModelRenderInfos *info = NULL;
    SCE_SListIterator *i = NULL;
    int ri = scene->ri;

    if (scene->render[ri].states.lighting)
    {
        SCE_List_ForEach (i, scene->render[ri].models)
        {
            info = SCE_List_GetData (i);
            /* uniquement si le modele a passe la validation pour etre rendu */
            if (info->draw)
                SCE_Scene_MakeLightsList (scene, SCE_List_GetData (i));
        }
    }
    else
    {
        SCE_List_ForEach (i, scene->render[ri].models)
        {
            info = SCE_List_GetData (i);
            info->num_lights = 0;
        }
    }
}

/* ajoute le 07/07/2008 */
void SCE_Scene_Update (SCE_SScene *scene, SCE_SCamera *camera,
                       SCE_STexture *rendertarget, SCEuint cubeface)
{
    int ri = scene->ri;

    /* assignation des variables "d'environnement" */
    scene->render[ri].rendertarget = rendertarget;
    scene->render[ri].cubeface = cubeface;
    scene->render[ri].camera = camera;

    SCE_Scene_SelectActivedModels (scene);

    /* --- gestion du frustum culling --- */
    if (scene->render[ri].states.frustum_culling)
        SCE_Scene_SelectInFrustumModels (scene);

    /* TODO: rajouter par la suite d'autres tris des objets */

    /* --- calcul de toutes les lumieres a portee des objets --- */
    SCE_Scene_MakeLightsLists (scene);
}

/* ajoute le 10/07/2008 */
void SCE_Scene_PushUpdate (SCE_SScene *scene)
{
    if (scene->ri + 1 >= SCE_MAX_RENDER_INFOS)
    {
        Logger_PrintMsg ("stack overflow in updates' push list of the scene");
        return;
    }
    scene->ri++;
}
/* ajoute le 10/07/2008 */
void SCE_Scene_PopUpdate (SCE_SScene *scene)
{
    if (scene->ri > 0)
        scene->ri--;
}

/* ajoute le 08/07/2008 */
static void SCE_Scene_ActivateLights (SCE_SSceneModelRenderInfos *infos)
{
    unsigned int i;
    for (i=0; i < infos->num_lights; i++)
        SCE_Light_Use (infos->lights[i]);
}

/* ajoute le 19/03/2008 */
/* revise le 09/07/2008 */
void SCE_Scene_Render (SCE_SScene *scene)
{
    SCE_SListIterator *i = NULL;
    SCE_SSceneModelRenderInfos *infos = NULL;
    int ri = scene->ri;
    SCE_SCamera *cam = scene->render[ri].camera;

    SCE_Scene_PrepareRendering (scene);

    /* mise en place du render target */
    SCE_Texture_RenderTo (scene->render[ri].rendertarget,
                          scene->render[ri].cubeface);

    /* activation de la camera et mise en place des matrices */
    SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
    SCE_CPushMatrix ();
    SCE_Camera_Use (cam);

    /* assignation de la matrice camera au gestionnaire de lumieres */
    if (scene->render[ri].states.lighting)
        SCE_Light_SetCameraMatrix (SCE_Camera_GetView (cam));
    /* pour chaque modele de la scene */
    SCE_List_ForEach (i, scene->render[ri].models)
    {
        infos = SCE_List_GetData (i);
        if (infos->draw)
        {
            SCE_Scene_ActivateLights (infos);
            SCE_Model_Render (infos->model);
            SCE_Light_Use (NULL);
        }
    }

    SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
    SCE_CPopMatrix ();

    SCE_CFlush ();
}
